/*
 * scpu/scpu.c
 *
 *      Execute programs written in a very simple machine language.
 *      See the course slides for the description of the machine
 *      language.
 */

#define _POSIX_C_SOURCE 201112L

#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <ctype.h>
#include <errno.h>

#define OPCODE_DATA     0x00
#define OPCODE_LOAD	0x20
#define OPCODE_STORE    0x40
#define OPCODE_ADD	0x60
#define OPCODE_SUB	0x80
#define OPCODE_EQUAL	0xa0
#define OPCODE_JUMP	0xc0
#define OPCODE_HALT	0xe0

#define OPMODE_IMM     	0x10
#define OPMODE_MEM	0x00

#define OPCODE_MASK	0xe0
#define OPERAND_MASK	0x0f
#define OPMODE_MASK	0x10

#define MEMSIZE		16

typedef struct {
    const char   *name;                 /* name of the program */
    unsigned char memory[MEMSIZE];      /* memory image of the program */
} program_t;

static const char *progname = "scpu";
static int dflag = 0;
static int tflag = 0;
static int qflag = 0;
static int wflag = -1;

typedef struct token {
    const char *label;
    unsigned char code;
} token_t;

static token_t token[] = {
    { .label = "LOAD",  .code = OPCODE_LOAD },
    { .label = "STORE", .code = OPCODE_STORE },
    { .label = "ADD",   .code = OPCODE_ADD },
    { .label = "SUB",   .code = OPCODE_SUB },
    { .label = "EQUAL", .code = OPCODE_EQUAL },
    { .label = "JUMP",  .code = OPCODE_JUMP },
    { .label = "HALT",  .code = OPCODE_HALT },
    { .label = "DATA",  .code = OPCODE_DATA },
    { .label = NULL,    .code = 0 }
};

static int
parse_line(char *line, program_t *prog)
{
    int i, no, op;
    char *to;
    unsigned char mode = 0;
    char *input = line;
    int len;
    
    if (sscanf(input, "%d:%n", &no, &len) != 1) {
        fprintf(stderr, "%s: missing label in '%s'\n", progname, line);
        return -1;
    }
    if (no < 0 || no >= MEMSIZE) {
	fprintf(stderr, "%s: illegal label in '%s'\n", progname, line);
        return -1;
    }
    input += len;

    /*
     * MacOs does not support %ms and hence we duplicate the input in
     * order to obtain a buffer large enough to hold %s. Apple could
     * do a better job with their POSIX support...
     */
    
    to = strdup(input);
    if (! to) {
        fprintf(stderr, "%s: memory allocation failure\n", progname);
        return -1;
    }
    if (sscanf(input, "%s%n", to, &len) != 1) {
        fprintf(stderr, "%s: missing instruction in '%s'\n", progname, line);
        return -1;
    }
    for (i = 0; token[i].label; i++) {
        if (strcmp(token[i].label, to) == 0) break;
    }
    free(to);
    if (! token[i].label) {
        fprintf(stderr, "%s: unknown instruction in '%s'\n", progname, line);
        return -1;
    }
    input += len;

    if (token[i].code != OPCODE_HALT) {
        if (sscanf(input, " #%d%n", &op, &len) == 1) {
            mode = OPMODE_IMM;
            if ((token[i].code != OPCODE_DATA && (op < 0 || op > 15))
                || (token[i].code == OPCODE_DATA && (op < 0 || op > 255))) {
                fprintf(stderr, "%s: illegal operand in '%s'\n", progname, line);
                return -1;
            }
        } else if (sscanf(input, " %d%n", &op, &len) == 1) {
            mode = OPMODE_MEM;
            if ((token[i].code != OPCODE_DATA && (op < 0 || op > 15))
                || (token[i].code == OPCODE_DATA && (op < 0 || op > 255))) {
                fprintf(stderr, "%s: illegal operand in '%s'\n", progname, line);
                return -1;
            }
        } else {
            fprintf(stderr, "%s: unknown operand in '%s'\n", progname, line);
            return -1;
        }
    }

    switch (token[i].code) {
    case OPCODE_HALT:
        prog->memory[no] = token[i].code;
        break;
    case OPCODE_DATA:
        prog->memory[no] = op;
        break;
    default:
        prog->memory[no] = token[i].code + op + mode;
    }        
    return 0;
}

static int
parse_file(const char *filename, program_t *prog)
{
    FILE *fp;
    char *line = NULL;
    size_t linecap = 0;
    ssize_t linelen;
    char *trimmed;

    if (! filename || strcmp(filename, "-") == 0) {
        clearerr(stdin);
        fp = stdin;
        prog->name = "stdin";
    } else {
        fp = fopen(filename, "r");
        if (! fp) {
            fprintf(stderr, "%s: failed to open '%s': %s\n",
                    progname, filename, strerror(errno));
            return -1;
	}
        prog->name = filename;
    }

    memset(prog->memory, 0, MEMSIZE);

    while ((linelen = getline(&line, &linecap, fp)) > 0) {
        int n = 0;
        if (linelen && line[linelen-1] == '\n') {
            line[linelen-1] = 0;
        }
        
        for (trimmed = line; isspace(*trimmed); trimmed++) ;
        if (trimmed[0] == ';' || trimmed[0] == 0) {
            continue;
        }
        n = parse_line(trimmed, prog);
        if (n == -1) {
            return -1;
        }
    }

    free(line);
    
    if (fflush(stdout) || ferror(stdout) || ferror(fp)) {
        fprintf(stderr, "%s: failed to read '%s': %s\n",
                progname, filename ? filename : "stdin" , strerror(errno));
        return -1;
    }

    if (fp != stdin) {
        (void) fclose(fp);
    }

    return 0;
}

static void
dump(unsigned char *memory)
{
    int i;
    
    printf("MEM = 0x");
    for (i = 0; i < MEMSIZE; i++) {
        printf("_%02x", (memory[i] & 0xff));
    }
    printf("\n");
}

static int
run(unsigned char *memory)
{
    unsigned char accumulator = 0;
    unsigned char val = 0;
    unsigned char opcode;
    unsigned char operand;
    unsigned char op;
    unsigned char immediate;
    char *opmode;
    unsigned short pc = 0;
    unsigned long cnt = 0;
    char comment[80];
    int halted = 0;
    
    while (! halted) {
	cnt++;
	val = memory[pc];
	opcode = (val & OPCODE_MASK);
	operand = (val & OPERAND_MASK);
	immediate = ((val & OPMODE_MASK) == OPMODE_IMM);
        opmode = immediate ? "#" : "";
        op = immediate ? operand : memory[operand];
	
        if (wflag == pc) {
            dump(memory);
        }
        if (tflag) {
            printf("%lu:\tPC=%02d\tIR=0x%02x", cnt, pc, val);
        }
        
	switch (opcode) {
	case OPCODE_LOAD:
            snprintf(comment, sizeof(comment), "LOAD  %s%d", opmode, operand);
            accumulator = op;
	    break;
	case OPCODE_STORE:
	    if (immediate) {
		printf("\t  EXCEPTION\t; illegal operand\n");
		return -1;
	    }
	    snprintf(comment, sizeof(comment), "STORE %s%d", opmode, operand);
	    memory[operand] = accumulator;
	    break;
	case OPCODE_ADD:
            snprintf(comment, sizeof(comment), "ADD   %s%d", opmode, operand);
            accumulator += op;
	    break;
	case OPCODE_SUB:
            snprintf(comment, sizeof(comment), "SUB   %s%d", opmode, operand);
            accumulator -= op;
	    break;
	case OPCODE_EQUAL:
            snprintf(comment, sizeof(comment), "EQUAL %s%d", opmode, operand);
            if (accumulator == operand) {
                pc = (pc + 1) % MEMSIZE;
            }
	    break;
	case OPCODE_JUMP:
            snprintf(comment, sizeof(comment), "JUMP  %s%d", opmode, operand);
            pc = (op % MEMSIZE) - 1;
	    break;
	case OPCODE_HALT:
	    snprintf(comment, sizeof(comment), "HALT");
	    halted = 1;
	    break;
	default:
	    printf("\t  EXCEPTION\t; illegal opcode 0x%02x\n", opcode);
	    return -1;
	}
        if (tflag) {
            printf("\t  ACC=0x%02x", accumulator);
            printf("\t; %s\n", comment);
        }
	pc = (pc + 1) % MEMSIZE;
    }
    return 0;
}

static int
execute(program_t *program)
{
    unsigned char memory[MEMSIZE];
    int status;
    
    memcpy(memory, program->memory, MEMSIZE);
    if (! qflag) {
        dump(memory);
    }
    status = run(memory);
    if (! qflag) {
        dump(memory);
    }
    
    return (status == 0) ? EXIT_SUCCESS : EXIT_FAILURE;
}

static void
disassemble(program_t *program)
{
    unsigned char val = 0;
    unsigned char opcode;
    unsigned char operand;
    unsigned char immediate;
    char *opmode;
    int i;

    if (! qflag) {
        printf(";; %s --\n;;\n", program->name);
    }
    for (i = 0; i < MEMSIZE; i++) {
	val = program->memory[i];
	opcode  = (val & OPCODE_MASK);
	operand = (val & OPERAND_MASK);
        immediate = (val & OPMODE_MASK) == OPMODE_IMM;
	opmode  = (immediate) ? "#" : "";
        printf("%d:\t", i);
	switch (opcode) {
	case OPCODE_LOAD:
	    printf("LOAD  %s%-2d", opmode, operand);
	    break;
	case OPCODE_STORE:
	    printf("STORE %s%-2d", opmode, operand);
	    break;
	case OPCODE_ADD:
	    printf("ADD   %s%-2d", opmode, operand);
	    break;
	case OPCODE_SUB:
	    printf("SUB   %s%-2d", opmode, operand);
	    break;
	case OPCODE_EQUAL:
	    printf("EQUAL %s%-2d", opmode, operand);
	    break;
	case OPCODE_JUMP:
	    printf("JUMP  %s%-2d", opmode, operand);
	    break;
	case OPCODE_HALT:
	    printf("HALT    ");
	    break;
	default:
	    printf("DATA  %-2d", operand);
	    break;
	}
        if (! qflag) {
            if (tflag) {
                printf("\t; 0x%02x = 0b_%d%d%d_%d_%d%d%d%d", val,
                       (val >> 7) & 0x01, (val >> 6) & 0x01,
                       (val >> 5) & 0x01, (val >> 4) & 0x01,
                       (val >> 3) & 0x01, (val >> 2) & 0x01,
                       (val >> 1) & 0x01, (val >> 0) & 0x01);
            }
            switch (opcode) {
            case OPCODE_LOAD:
                if (immediate) {
                    printf("\t; acc = %d", operand);
                } else {
                    printf("\t; acc = m[%d]", operand);
                }
                break;
            case OPCODE_STORE:
                if (immediate) {
                    printf("\t; %d = acc", operand);
                } else {
                    printf("\t; m[%d] = acc", operand);
                }
                break;
            case OPCODE_ADD:
                if (immediate) {
                    printf("\t; acc += %d", operand);
                } else {
                    printf("\t; acc += m[%d]", operand);
                }
                break;
            case OPCODE_SUB:
                if (immediate) {
                    printf("\t; acc -= %d", operand);
                } else {
                    printf("\t; acc -= m[%d]", operand);
                }
                break;
            case OPCODE_EQUAL:
                if (immediate) {
                    printf("\t; if acc == %d pc++", operand);
                } else {
                    printf("\t; if acc == m[%d] pc++", operand);
                }
                break;
            case OPCODE_JUMP:
                if (immediate) {
                    printf("\t; pc = %d", operand);
                } else {
                    printf("\t; pc = m[%d]", operand);
                }
                break;
            case OPCODE_HALT:
                printf("\t; halt");
                break;
            default:
                printf("\t;");
                break;
            }
        }
        printf("\n");
    }
}

int
main(int argc, char **argv)
{
    int status = EXIT_SUCCESS;
    int c;
    program_t prog;

    while ((c = getopt(argc, argv, "dtqw:h")) >= 0) {
	switch (c) {
	case 'd':
	    dflag = 1;
	    break;
        case 't':
            tflag = 1;
            break;
        case 'q':
            qflag = 1;
            break;
        case 'w':
            wflag = atoi(optarg);
            break;
        case 'h':
            printf("usage: %s [-d] [-t] [-q] [-w pc] prog\n", progname);
            return(EXIT_SUCCESS);
	case ':':
	case '?':
	    exit(EXIT_FAILURE);
	}
    }

    if (optind == argc) {
        if (parse_file(NULL, &prog) < 0) {
            return EXIT_FAILURE;
	}
        if (dflag) {
            disassemble(&prog);
        } else {
            if (execute(&prog)) {
                status = EXIT_FAILURE;
            }
        }
    } else {
	for (; optind < argc; optind++) {
            if (parse_file(argv[optind], &prog) < 0) {
		status = EXIT_FAILURE;
                continue;
            }
            if (dflag) {
                disassemble(&prog);
            } else {
                if (execute(&prog)) {
                    status = EXIT_FAILURE;
                }
            }
        }
    }
    
    return status;
}
